from copy import deepcopy
from logging import getLogger

import joblib
import matplotlib.pyplot as plt
import numpy as np
import sklearn
import xarray as xr
from replay_trajectory_classification.bins import atleast_2d, get_centers
from replay_trajectory_classification.continuous_state_transitions import (
    EmpiricalMovement, RandomWalk, Uniform)
from replay_trajectory_classification.core import (_acausal_classify,
                                                   _acausal_classify_gpu,
                                                   _causal_classify,
                                                   _causal_classify_gpu, mask,
                                                   scaled_likelihood)
from replay_trajectory_classification.discrete_state_transitions import \
    DiagonalDiscrete
from replay_trajectory_classification.environments import Environment
from replay_trajectory_classification.initial_conditions import \
    UniformInitialConditions
from replay_trajectory_classification.likelihoods import (
    _ClUSTERLESS_ALGORITHMS, estimate_place_fields,
    estimate_spiking_likelihood)
from replay_trajectory_classification.misc import NumbaKDE
from replay_trajectory_classification.observation_model import ObservationModel
from sklearn.base import BaseEstimator

logger = getLogger(__name__)

sklearn.set_config(print_changed_only=False)

_DEFAULT_CLUSTERLESS_MODEL_KWARGS = {
    'model': NumbaKDE,
    'model_kwargs': {
        'bandwidth': np.array([24.0, 24.0, 24.0, 24.0, 6.0, 6.0])
    }
}

_DEFAULT_CONTINUOUS_TRANSITIONS = (
    [[RandomWalk(), Uniform()],
     [Uniform(),    Uniform()]])

_DEFAULT_ENVIRONMENT = Environment(environment_name='')


class _ClassifierBase(BaseEstimator):
    def __init__(self,
                 environments=_DEFAULT_ENVIRONMENT,
                 observation_models=None,
                 continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
                 discrete_transition_type=DiagonalDiscrete(0.968),
                 initial_conditions_type=UniformInitialConditions(),
                 infer_track_interior=True):
        if isinstance(environments, Environment):
            environments = (environments,)
        if observation_models is None:
            n_states = len(continuous_transition_types)
            env_name = environments[0].environment_name
            observation_models = (ObservationModel(env_name),) * n_states
        self.environments = environments
        self.observation_models = observation_models
        self.continuous_transition_types = continuous_transition_types
        self.discrete_transition_type = discrete_transition_type
        self.initial_conditions_type = initial_conditions_type
        self.infer_track_interior = infer_track_interior

    def fit_environments(self, position, environment_labels=None):
        for environment in self.environments:
            if environment_labels is None:
                is_environment = np.ones((position.shape[0],), dtype=bool)
            else:
                is_environment = (environment_labels ==
                                  environment.environment_name)
            environment.fit_place_grid(
                position[is_environment],
                infer_track_interior=self.infer_track_interior)

        self.max_pos_bins_ = np.max([env.place_bin_centers_.shape[0]
                                     for env in self.environments])

    def fit_initial_conditions(self):
        logger.info('Fitting initial conditions...')
        environment_names_to_state = [
            obs.environment_name for obs in self.observation_models]
        n_states = len(self.observation_models)
        initial_conditions = (
            self.initial_conditions_type.make_initial_conditions(
                self.environments, environment_names_to_state))

        self.initial_conditions_ = np.zeros((n_states, self.max_pos_bins_, 1),
                                            dtype=np.float64)
        for state_ind, ic in enumerate(initial_conditions):
            self.initial_conditions_[state_ind,
                                     :ic.shape[0]] = ic[..., np.newaxis]

    def fit_continuous_state_transition(
            self,
            continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
            position=None,
            is_training=None,
            encoding_group_labels=None,
            environment_labels=None,
    ):
        logger.info('Fitting state transition...')

        if is_training is None:
            n_time = position.shape[0]
            is_training = np.ones((n_time,), dtype=bool)

        if encoding_group_labels is None:
            n_time = position.shape[0]
            encoding_group_labels = np.zeros((n_time,), dtype=np.int32)

        is_training = np.asarray(is_training).squeeze()

        self.continuous_transition_types = continuous_transition_types
        continuous_state_transition = []

        for row in self.continuous_transition_types:
            continuous_state_transition.append([])
            for transition in row:
                if isinstance(transition, EmpiricalMovement):
                    continuous_state_transition[-1].append(
                        transition.make_state_transition(
                            self.environments, position, is_training,
                            encoding_group_labels, environment_labels))
                else:
                    continuous_state_transition[-1].append(
                        transition.make_state_transition(self.environments))

        n_states = len(self.continuous_transition_types)
        self.continuous_state_transition_ = np.zeros(
            (n_states, n_states, self.max_pos_bins_, self.max_pos_bins_))

        for row_ind, row in enumerate(continuous_state_transition):
            for column_ind, st in enumerate(row):
                self.continuous_state_transition_[
                    row_ind, column_ind, :st.shape[0], :st.shape[1]] = st

    def fit_discrete_state_transition(self):
        n_states = len(self.continuous_transition_types)
        self.discrete_state_transition_ = (
            self.discrete_transition_type.make_state_transition(n_states))

    def _get_results(self, likelihood, n_time, time, state_names, use_gpu,
                     is_compute_acausal):
        n_states = self.discrete_state_transition_.shape[0]
        results = {}

        if use_gpu:
            results['likelihood'] = np.full(
                (n_time, n_states, self.max_pos_bins_, 1), np.nan,
                dtype=np.float32)
        else:
            results['likelihood'] = np.full(
                (n_time, n_states, self.max_pos_bins_, 1), np.nan,
                dtype=np.float64)

        for state_ind, obs in enumerate(self.observation_models):
            likelihood_name = (obs.environment_name, obs.encoding_group)
            n_bins = likelihood[likelihood_name].shape[1]
            results['likelihood'][:, state_ind, :n_bins] = (
                likelihood[likelihood_name][..., np.newaxis])

        results['likelihood'] = scaled_likelihood(
            results['likelihood'], axis=(1, 2))
        results['likelihood'][np.isnan(results['likelihood'])] = 0.0

        n_environments = len(self.environments)
        if n_environments == 1:
            logger.info('Estimating causal posterior...')
            is_track_interior = (self.environments[0]
                                 .is_track_interior_.ravel(order='F'))
            n_position_bins = len(is_track_interior)
            is_states = np.ones((n_states,), dtype=bool)
            st_interior_ind = np.ix_(
                is_states, is_states, is_track_interior, is_track_interior)
            if not use_gpu:
                results['causal_posterior'] = np.full(
                    (n_time, n_states, n_position_bins, 1), np.nan,
                    dtype=np.float64)
                (results['causal_posterior'][:, :, is_track_interior],
                 data_log_likelihood) = (
                    _causal_classify(
                        self.initial_conditions_[:, is_track_interior],
                        self.continuous_state_transition_[st_interior_ind],
                        self.discrete_state_transition_,
                        results['likelihood'][:, :, is_track_interior]))
            else:
                results['causal_posterior'] = np.full(
                    (n_time, n_states, n_position_bins, 1), np.nan,
                    dtype=np.float32)
                (results['causal_posterior'][:, :, is_track_interior],
                 data_log_likelihood) = (
                    _causal_classify_gpu(
                        self.initial_conditions_[:, is_track_interior],
                        self.continuous_state_transition_[st_interior_ind],
                        self.discrete_state_transition_,
                        results['likelihood'][:, :, is_track_interior]))

            if is_compute_acausal:
                logger.info('Estimating acausal posterior...')
                if not use_gpu:
                    results['acausal_posterior'] = np.full(
                        (n_time, n_states, n_position_bins, 1), np.nan,
                        dtype=np.float64)
                    results['acausal_posterior'][:, :, is_track_interior] = (
                        _acausal_classify(
                            results['causal_posterior'][:,
                                                        :, is_track_interior],
                            self.continuous_state_transition_[st_interior_ind],
                            self.discrete_state_transition_))
                else:
                    results['acausal_posterior'] = np.full(
                        (n_time, n_states, n_position_bins, 1), np.nan,
                        dtype=np.float32)
                    results['acausal_posterior'][:, :, is_track_interior] = (
                        _acausal_classify_gpu(
                            results['causal_posterior'][:,
                                                        :, is_track_interior],
                            self.continuous_state_transition_[st_interior_ind],
                            self.discrete_state_transition_))

            if time is None:
                time = np.arange(n_time)

            return self._convert_results_to_xarray(
                results, time, state_names, data_log_likelihood)

        else:
            logger.info('Estimating causal posterior...')
            if not use_gpu:
                (results['causal_posterior'],
                 data_log_likelihood) = _causal_classify(
                    self.initial_conditions_,
                    self.continuous_state_transition_,
                    self.discrete_state_transition_,
                    results['likelihood'])
            else:
                (results['causal_posterior'],
                 data_log_likelihood) = _causal_classify_gpu(
                    self.initial_conditions_,
                    self.continuous_state_transition_,
                    self.discrete_state_transition_,
                    results['likelihood'])

            if is_compute_acausal:
                logger.info('Estimating acausal posterior...')
                if not use_gpu:
                    results['acausal_posterior'] = _acausal_classify(
                        results['causal_posterior'],
                        self.continuous_state_transition_,
                        self.discrete_state_transition_)
                else:
                    results['acausal_posterior'] = _acausal_classify_gpu(
                        results['causal_posterior'],
                        self.continuous_state_transition_,
                        self.discrete_state_transition_)

            if time is None:
                time = np.arange(n_time)

            return self._convert_results_to_xarray_mutienvironment(
                results, time, state_names, data_log_likelihood)

    def _convert_results_to_xarray(self, results, time, state_names,
                                   data_log_likelihood):
        attrs = {'data_log_likelihood': data_log_likelihood}
        n_position_dims = self.environments[0].place_bin_centers_.shape[1]
        diag_transition_names = np.diag(
            np.asarray(self.continuous_transition_types))
        if state_names is None:
            if len(np.unique(self.observation_models)) == 1:
                state_names = diag_transition_names
            else:
                state_names = [
                    f"{obs.encoding_group}-{transition}" for obs, transition
                    in zip(self.observation_models,
                           diag_transition_names)]
        n_time = time.shape[0]
        n_states = len(state_names)
        is_track_interior = (
            self.environments[0].is_track_interior_.ravel(order='F'))
        edges = self.environments[0].edges_

        if n_position_dims > 1:
            centers_shape = self.environments[0].centers_shape_
            new_shape = (n_time, n_states, *centers_shape)
            dims = ['time', 'state', 'x_position', 'y_position']
            coords = dict(
                time=time,
                x_position=get_centers(edges[0]),
                y_position=get_centers(edges[1]),
                state=state_names,
            )
            results = xr.Dataset(
                {key: (dims,
                       (mask(value, is_track_interior).squeeze(axis=-1)
                        .reshape(new_shape).swapaxes(-1, -2)))
                 for key, value in results.items()},
                coords=coords,
                attrs=attrs)
        else:
            dims = ['time', 'state', 'position']
            coords = dict(
                time=time,
                position=get_centers(edges[0]),
                state=state_names,
            )
            results = xr.Dataset(
                {key: (dims, (mask(value, is_track_interior).squeeze(axis=-1)))
                 for key, value in results.items()},
                coords=coords,
                attrs=attrs)

        return results

    def _convert_results_to_xarray_mutienvironment(self, results, time,
                                                   state_names,
                                                   data_log_likelihood):
        if state_names is None:
            state_names = [f'{obs.environment_name}-{obs.encoding_group}'
                           for obs in self.observation_models]

        attrs = {'data_log_likelihood': data_log_likelihood}
        n_position_dims = self.environments[0].place_bin_centers_.shape[1]

        if n_position_dims > 1:
            dims = ['time', 'state', 'position']
            coords = dict(
                time=time,
                state=state_names,
            )
            for env in self.environments:
                coords[env.environment_name +
                       '_x_position'] = get_centers(env.edges_[0])
                coords[env.environment_name +
                       '_y_position'] = get_centers(env.edges_[1])
            results = xr.Dataset(
                {key: (dims, value.squeeze(axis=-1))
                 for key, value in results.items()},
                coords=coords,
                attrs=attrs)
        else:
            dims = ['time', 'state', 'position']
            coords = dict(
                time=time,
                state=state_names,
            )
            for env in self.environments:
                coords[env.environment_name +
                       '_position'] = get_centers(env.edges_[0])

            results = xr.Dataset(
                {key: (dims, value.squeeze(axis=-1))
                 for key, value in results.items()},
                coords=coords,
                attrs=attrs)

        return results

    def fit(self):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def save_model(self, filename='model.pkl'):
        joblib.dump(self, filename)

    @staticmethod
    def load_model(filename='model.pkl'):
        return joblib.load(filename)

    @staticmethod
    def predict_proba(results):
        try:
            return results.sum(['x_position', 'y_position'])
        except ValueError:
            return results.sum(['position'])

    def copy(self):
        return deepcopy(self)


class SortedSpikesClassifier(_ClassifierBase):
    '''

    Attributes
    ----------
    environment : Environment
        The spatial environment and topology to fit
    continuous_transition_types : tuple of tuples
        The continuous state transition class instances to fit
    discrete_transition_type : Discrete
        The type of transition between states
    initial_conditions_type : InitialConditions
        The initial conditions class instance to fit
    infer_track_interior : bool, optional
        Whether to infer the valid position bins
    knot_spacing : float, optional
        How far apart the spline knots are in position
    spike_model_penalty : float, optional
        L2 penalty (ridge) for the size of the regression coefficients

    '''

    def __init__(self,
                 environments=_DEFAULT_ENVIRONMENT,
                 observation_models=None,
                 continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
                 discrete_transition_type=DiagonalDiscrete(0.98),
                 initial_conditions_type=UniformInitialConditions(),
                 infer_track_interior=True,
                 knot_spacing=10,
                 spike_model_penalty=1E1):
        super().__init__(environments,
                         observation_models,
                         continuous_transition_types,
                         discrete_transition_type,
                         initial_conditions_type,
                         infer_track_interior)
        self.knot_spacing = knot_spacing
        self.spike_model_penalty = spike_model_penalty

    def fit_place_fields(self,
                         position,
                         spikes,
                         is_training=None,
                         encoding_group_labels=None,
                         environment_labels=None):
        logger.info('Fitting place fields...')
        n_time = position.shape[0]
        if is_training is None:
            is_training = np.ones((n_time,), dtype=bool)

        if encoding_group_labels is None:
            encoding_group_labels = np.zeros((n_time,), dtype=np.int32)

        if environment_labels is None:
            environment_labels = np.asarray(
                [self.environments[0].environment_name] * n_time)

        is_training = np.asarray(is_training).squeeze()

        self.place_fields_ = {}
        for obs in np.unique(self.observation_models):
            environment = self.environments[
                self.environments.index(obs.environment_name)]

            is_encoding = np.isin(encoding_group_labels, obs.encoding_group)
            is_environment = (environment_labels == obs.environment_name)
            likelihood_name = (obs.environment_name, obs.encoding_group)

            self.place_fields_[likelihood_name] = estimate_place_fields(
                position=position[is_training & is_encoding & is_environment],
                spikes=spikes[is_training & is_encoding & is_environment],
                place_bin_centers=environment.place_bin_centers_,
                place_bin_edges=environment.place_bin_edges_,
                penalty=self.spike_model_penalty,
                knot_spacing=self.knot_spacing)

    def plot_place_fields(self, sampling_frequency=1, figsize=(10, 7)):
        try:
            for (env, enc) in self.place_fields_:
                is_track_interior = self.environments[self.environments.index(
                    env)].is_track_interior_[np.newaxis]
                ((self.place_fields_[(env, enc)] * sampling_frequency)
                 .unstack('position')
                 .where(is_track_interior)
                 .plot(x='x_position', y='y_position',  col='neuron',
                       col_wrap=8, vmin=0.0, vmax=3.0))
        except ValueError:
            fig, axes = plt.subplots(
                len(self.place_fields_), 1, constrained_layout=True,
                figsize=figsize)
            for ax, ((env_name, enc_group), place_fields) in zip(
                    axes.flat, self.place_fields_.items()):
                is_track_interior = self.environments[self.environments.index(
                    env_name)].is_track_interior_[:, np.newaxis]
                ((place_fields * sampling_frequency)
                 .where(is_track_interior)
                 .plot(
                    x='position', hue='neuron', add_legend=False, ax=ax))
                ax.set_title(
                    f'Environment = {env_name}, Encoding Group = {enc_group}')
                ax.set_ylabel('Firing Rate\n[spikes/s]')

    def fit(self,
            position,
            spikes,
            is_training=None,
            encoding_group_labels=None,
            environment_labels=None,
            ):
        '''

        Parameters
        ----------
        position : ndarray, shape (n_time, n_position_dims)
        spikes : ndarray, shape (n_time, n_neurons)
        is_training : None or bool ndarray, shape (n_time), optional
            Time bins to be used for encoding.
        encoding_group_labels : None or ndarray, shape (n_time,)
            Label for the corresponding encoding group for each time point
        environment_labels : None or ndarray, shape (n_time,)
            Label for the corresponding environment for each time point

        Returns
        -------
        self

        '''
        position = atleast_2d(np.asarray(position))
        spikes = np.asarray(spikes)
        self.fit_environments(position, environment_labels)
        self.fit_initial_conditions()
        self.fit_continuous_state_transition(
            self.continuous_transition_types,
            position,
            is_training,
            encoding_group_labels,
            environment_labels,
        )
        self.fit_discrete_state_transition()
        self.fit_place_fields(position,
                              spikes,
                              is_training,
                              encoding_group_labels,
                              environment_labels)

        return self

    def predict(self, spikes, time=None, is_compute_acausal=True,
                use_gpu=False,
                state_names=None):
        '''

        Parameters
        ----------
        spikes : ndarray, shape (n_time, n_neurons)
        time : ndarray or None, shape (n_time,), optional
        is_compute_acausal : bool, optional
        use_gpu : bool, optional
            Use GPU for the state space part of the model, not the likelihood.
        state_names : None or array_like, shape (n_states,)

        Returns
        -------
        results : xarray.Dataset

        '''
        spikes = np.asarray(spikes)
        n_time = spikes.shape[0]

        # likelihood
        logger.info('Estimating likelihood...')
        likelihood = {}
        for (env_name, enc_group), place_fields in self.place_fields_.items():
            env_ind = self.environments.index(env_name)
            is_track_interior = (
                self.environments[env_ind].is_track_interior_.ravel(order='F'))
            likelihood[(env_name, enc_group)] = estimate_spiking_likelihood(
                spikes,
                place_fields.values,
                is_track_interior)

        return self._get_results(
            likelihood, n_time, time, state_names, use_gpu, is_compute_acausal)


class ClusterlessClassifier(_ClassifierBase):
    '''

    Attributes
    ----------
    environment : Environment
        The spatial environment and topology to fit
    continuous_transition_types : tuple of tuples
        The continuous state transition class instances to fit
    discrete_transition_type : Discrete
        The type of transition between states
    initial_conditions_type : InitialConditions
        The initial conditions class instance to fit
    infer_track_interior : bool, optional
        Whether to infer the valid position bins
    clusterless_algorithm : str
        The type of clusterless algorithm. See _ClUSTERLESS_ALGORITHMS for keys
    clusterless_algorithm_params : dict
        Parameters for the clusterless algorithms.

    '''

    def __init__(self,
                 environments=_DEFAULT_ENVIRONMENT,
                 observation_models=None,
                 continuous_transition_types=_DEFAULT_CONTINUOUS_TRANSITIONS,
                 discrete_transition_type=DiagonalDiscrete(0.98),
                 initial_conditions_type=UniformInitialConditions(),
                 infer_track_interior=True,
                 clusterless_algorithm='multiunit_likelihood',
                 clusterless_algorithm_params=_DEFAULT_CLUSTERLESS_MODEL_KWARGS
                 ):
        super().__init__(environments,
                         observation_models,
                         continuous_transition_types,
                         discrete_transition_type,
                         initial_conditions_type,
                         infer_track_interior)

        self.clusterless_algorithm = clusterless_algorithm
        self.clusterless_algorithm_params = clusterless_algorithm_params

    def fit_multiunits(self,
                       position,
                       multiunits,
                       is_training=None,
                       encoding_group_labels=None,
                       environment_labels=None):
        '''

        Parameters
        ----------
        position : array_like, shape (n_time, n_position_dims)
        multiunits : array_like, shape (n_time, n_marks, n_electrodes)
        is_training : None or array_like, shape (n_time,)
        encoding_group_labels : None or ndarray, shape (n_time,)
            Label for the corresponding encoding group for each time point
        environment_labels : None or ndarray, shape (n_time,)
            Label for the corresponding environment for each time point

        '''
        logger.info('Fitting multiunits...')
        n_time = position.shape[0]
        if is_training is None:
            is_training = np.ones((n_time,), dtype=bool)

        if encoding_group_labels is None:
            encoding_group_labels = np.zeros((n_time,), dtype=np.int32)

        if environment_labels is None:
            environment_labels = np.asarray(
                [self.environments[0].environment_name] * n_time)

        is_training = np.asarray(is_training).squeeze()

        kwargs = self.clusterless_algorithm_params
        if kwargs is None:
            kwargs = {}

        self.encoding_model_ = {}

        for obs in np.unique(self.observation_models):
            environment = self.environments[
                self.environments.index(obs.environment_name)]

            is_encoding = np.isin(encoding_group_labels, obs.encoding_group)
            is_environment = (environment_labels == obs.environment_name)
            is_group = is_training & is_encoding & is_environment

            likelihood_name = (obs.environment_name, obs.encoding_group)

            self.encoding_model_[likelihood_name] = _ClUSTERLESS_ALGORITHMS[
                self.clusterless_algorithm][0](
                    position=position[is_group],
                    multiunits=multiunits[is_group],
                    place_bin_centers=environment.place_bin_centers_,
                    is_track_interior=environment.is_track_interior_.ravel(
                        order='F'),
                    **kwargs
            )

    def fit(self,
            position,
            multiunits,
            is_training=None,
            encoding_group_labels=None,
            environment_labels=None,
            ):
        '''

        Parameters
        ----------
        position : array_like, shape (n_time, n_position_dims)
        multiunits : array_like, shape (n_time, n_marks, n_electrodes)
        is_training : None or array_like, shape (n_time,)
        encoding_group_labels : None or ndarray, shape (n_time,)
            Label for the corresponding encoding group for each time point
        environment_labels : None or ndarray, shape (n_time,)
            Label for the corresponding environment for each time point

        Returns
        -------
        self

        '''
        position = atleast_2d(np.asarray(position))
        multiunits = np.asarray(multiunits)

        self.fit_environments(position, environment_labels)
        self.fit_initial_conditions()
        self.fit_continuous_state_transition(
            self.continuous_transition_types,
            position,
            is_training,
            encoding_group_labels,
            environment_labels,
        )
        self.fit_discrete_state_transition()
        self.fit_multiunits(position,
                            multiunits,
                            is_training,
                            encoding_group_labels,
                            environment_labels)

        return self

    def predict(self, multiunits, time=None, is_compute_acausal=True,
                use_gpu=False, state_names=None):
        '''

        Parameters
        ----------
        multiunits : array_like, shape (n_time, n_marks, n_electrodes)
        time : None or ndarray, shape (n_time,)
        is_compute_acausal : bool, optional
            Use future information to compute the posterior.
        use_gpu : bool, optional
            Use GPU for the state space part of the model, not the likelihood.
        state_names : None or array_like, shape (n_states,)

        Returns
        -------
        results : xarray.Dataset

        '''
        multiunits = np.asarray(multiunits)
        n_time = multiunits.shape[0]

        logger.info('Estimating likelihood...')
        likelihood = {}
        for (env_name, enc_group), encoding_params in self.encoding_model_.items():
            env_ind = self.environments.index(env_name)
            is_track_interior = (
                self.environments[env_ind].is_track_interior_.ravel(order='F'))
            place_bin_centers = self.environments[env_ind].place_bin_centers_
            likelihood[(env_name, enc_group)] = _ClUSTERLESS_ALGORITHMS[
                self.clusterless_algorithm][1](
                    multiunits=multiunits,
                    place_bin_centers=place_bin_centers,
                    is_track_interior=is_track_interior,
                    **encoding_params
            )

        return self._get_results(
            likelihood, n_time, time, state_names, use_gpu, is_compute_acausal)



import math

import numpy as np
from numba import cuda
from numba.types import float32
from replay_trajectory_classification.bins import atleast_2d

# Precompute this constant as a float32.  Numba will inline it at compile time.
SQRT_2PI = np.float32((2 * math.pi)**0.5)
EPS = np.spacing(1)


@cuda.jit(device=True)
def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


@cuda.jit
def kde(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_bandwidths = len(bandwidths)
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    for thread_id in range(cuda.grid(1), n_eval_points, cuda.gridsize(1)):
        sum_kernel = float32(0.0)
        for sample_ind in range(n_samples):
            product_kernel = float32(1.0)
            for bandwidth_ind in range(n_bandwidths):
                product_kernel *= gaussian_pdf(
                    eval_points[thread_id, bandwidth_ind],
                    samples[sample_ind, bandwidth_ind],
                    bandwidths[bandwidth_ind]
                )
            sum_kernel += product_kernel

        out[thread_id] = sum_kernel / n_samples


def get_marks_by_place_bin_centers(marks, place_bin_centers):
    """

    Parameters
    ----------
    marks : ndarray, shape (n_spikes, n_features)
    place_bin_centers : ndarray, shape (n_position_bins, n_position_dims)

    Returns
    -------
    marks_by_place_bin_centers : ndarray, shape (n_spikes * n_position_bins,
                                                 n_features + n_position_dims)

    """
    n_spikes = marks.shape[0]
    n_place_bin_centers = place_bin_centers.shape[0]
    return np.concatenate(
        (np.tile(marks, reps=(n_place_bin_centers, 1)),
         np.repeat(place_bin_centers, n_spikes, axis=0)), axis=1)


def estimate_log_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.log(mean_rate) + np.log(density) - np.log(occupancy)


def estimate_pdf(
    decoding_marks, encoding_marks, mark_std,
    place_bin_centers, encoding_positions, position_std, stream=0
):
    '''

    Parameters
    ----------
    decoding_marks : ndarray, shape (n_decoding_spikes, n_marks)
    encoding_marks : ndarray, shape (n_encoding_spikes, n_marks)
    mark_std : float
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : ndarray, shape (n_encoding_spikes, n_position_dims)
    position_std : float
    stream : numba.cuda.stream, optional

    Returns
    -------
    joint_mark_intensity : ndarray, shape (n_decoding_spikes, n_bins)

    '''
    decoding_marks = np.atleast_2d(decoding_marks)

    # Copy the arrays to the GPU
    eval_points = cuda.to_device(
        get_marks_by_place_bin_centers(decoding_marks, place_bin_centers)
        .astype(np.float32),
        stream=stream)
    encoding_samples = cuda.to_device(
        np.concatenate((encoding_marks, encoding_positions), axis=1)
        .astype(np.float32), stream=stream)

    n_marks = decoding_marks.shape[1]
    n_position_dims = place_bin_centers.shape[1]
    bandwidths = cuda.to_device(
        np.concatenate(
            ([mark_std] * n_marks,
             [position_std] * n_position_dims,
             )
        ).astype(np.float32), stream=stream)
    n_eval_points = len(eval_points)
    # Allocate memory on the GPU for the result
    pdf = cuda.device_array(
        shape=(n_eval_points,), dtype=np.float32, stream=stream)

    # Run KDE
    kde.forall(n_eval_points, stream=stream)(
        eval_points, encoding_samples, bandwidths, pdf)

    return pdf


def estimate_multiunit_likelihood_gpu(multiunits,
                                      encoding_marks,
                                      mark_std,
                                      place_bin_centers,
                                      encoding_positions,
                                      position_std,
                                      occupancy,
                                      mean_rates,
                                      summed_ground_process_intensity,
                                      is_track_interior=None,
                                      time_bin_size=1,
                                      n_streams=2):
    '''

    Parameters
    ----------
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    encoding_marks : list of ndarrays, len (n_electrodes,)
    mark_std : float
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : list of ndarrays, len (n_electrodes,)
    position_std : float
    occupancy : ndarray, (n_bins,)
    mean_rates : list, len (n_electrodes,)
    summed_ground_process_intensity : ndarray, shape (n_bins,)

    Returns
    -------
    log_likelihood : ndarray, shape (n_time, n_bins)

    '''

    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)

    n_time = multiunits.shape[0]
    log_likelihood = (-time_bin_size * summed_ground_process_intensity *
                      np.ones((n_time, 1)))
    n_electrodes = multiunits.shape[-1]
    multiunits = np.moveaxis(multiunits, -1, 0)
    streams = [cuda.stream() for _ in range(min(n_streams, n_electrodes))]
    pdfs = []
    is_spikes = []

    for elec_ind, (multiunit, enc_marks, enc_pos) in enumerate(zip(
            multiunits, encoding_marks, encoding_positions)):
        nan_multiunit = np.isnan(multiunit)
        is_spike = np.any(~nan_multiunit, axis=1)
        nan_mark_dims = np.all(nan_multiunit, axis=0)

        is_spikes.append(is_spike)
        n_spikes = is_spike.sum()
        if n_spikes > 0:
            pdfs.append(estimate_pdf(
                multiunit[np.ix_(is_spike, ~nan_mark_dims)],
                enc_marks,
                mark_std,
                place_bin_centers[is_track_interior],
                enc_pos,
                position_std,
                stream=streams[elec_ind % n_streams]
            ))
        else:
            pdfs.append([])

    n_interior_place_bins = is_track_interior.sum()
    for elec_ind, (pdf, mean_rate, is_spike) in enumerate(
            zip(pdfs, mean_rates, is_spikes)):
        n_spikes = is_spike.sum()
        if n_spikes > 0:
            # Copy results from GPU to CPU and
            # reshape to (n_decoding_spikes, n_interior_place_bins)
            pdf = (pdf
                   .copy_to_host(stream=streams[elec_ind % n_streams])
                   .reshape((n_spikes, n_interior_place_bins), order='F'))
            log_intensity = estimate_log_intensity(
                pdf,
                occupancy[is_track_interior],
                mean_rate)
            is_inf = np.all(np.isneginf(log_intensity), axis=1)
            log_intensity[is_inf] = np.spacing(1)
            log_likelihood[np.ix_(
                is_spike, is_track_interior)] += log_intensity

    log_likelihood[:, ~is_track_interior] = np.nan

    return log_likelihood


def estimate_position_density(place_bin_centers, positions, position_std):
    # Copy the arrays to the device
    eval_points = cuda.to_device(place_bin_centers.astype(np.float32))
    samples = cuda.to_device(positions.astype(np.float32))
    n_position_dims = positions.shape[1]
    bandwidths = cuda.to_device(
        np.asarray([position_std] * n_position_dims).astype(np.float32))

    # Allocate memory on the device for the result
    n_eval_points = len(eval_points)
    out = cuda.device_array((n_eval_points,), dtype=np.float32)

    kde.forall(n_eval_points)(
        eval_points, samples, bandwidths, out)

    return out.copy_to_host()


def estimate_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.exp(estimate_log_intensity(density, occupancy, mean_rate))


def fit_multiunit_likelihood_gpu(position,
                                 multiunits,
                                 place_bin_centers,
                                 mark_std,
                                 position_std,
                                 is_track_interior=None,
                                 **kwargs):
    '''

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, shape ( n_bins, n_position_dims)
    mark_std : float
    position_std : float
    is_track_interior : None or ndarray, shape (n_bins,)

    Returns
    -------
    encoding_model : dict

    '''
    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)
    position = atleast_2d(position)
    place_bin_centers = atleast_2d(place_bin_centers)

    not_nan_position = np.all(~np.isnan(position), axis=1)

    occupancy = np.zeros((place_bin_centers.shape[0],), dtype=np.float32)
    occupancy[is_track_interior] = estimate_position_density(
        place_bin_centers[is_track_interior],
        position[not_nan_position],
        position_std)

    mean_rates = []
    ground_process_intensities = []
    encoding_marks = []
    encoding_positions = []

    for multiunit in np.moveaxis(multiunits, -1, 0):

        # ground process intensity
        nan_multiunit = np.isnan(multiunit)
        is_spike = np.any(~nan_multiunit, axis=1)
        nan_mark_dims = np.all(nan_multiunit, axis=0)

        mean_rates.append(is_spike.mean())
        marginal_density = np.zeros(
            (place_bin_centers.shape[0],), dtype=np.float32)

        if is_spike.sum() > 0:
            marginal_density[is_track_interior] = estimate_position_density(
                place_bin_centers[is_track_interior],
                position[is_spike & not_nan_position], position_std)

        ground_process_intensities.append(
            estimate_intensity(marginal_density, occupancy, mean_rates[-1])
            + EPS)

        encoding_marks.append(
            multiunit[np.ix_(is_spike & not_nan_position, ~nan_mark_dims)
                      ].astype(np.float32))
        encoding_positions.append(position[is_spike & not_nan_position])

    summed_ground_process_intensity = np.sum(
        np.stack(ground_process_intensities, axis=0), axis=0, keepdims=True)

    return {
        'encoding_marks': encoding_marks,
        'encoding_positions': encoding_positions,
        'summed_ground_process_intensity': summed_ground_process_intensity,
        'occupancy': occupancy,
        'mean_rates': mean_rates,
        'mark_std': mark_std,
        'position_std': position_std,
        **kwargs,
    }



# flake8: noqa
from replay_trajectory_classification.classifier import (
    ClusterlessClassifier, SortedSpikesClassifier)
from replay_trajectory_classification.decoder import (ClusterlessDecoder,
                                                      SortedSpikesDecoder)
from track_linearization import (get_linearized_position,
                                 make_actual_vs_linearized_position_movie,
                                 make_track_graph, plot_graph_as_1D,
                                 plot_track_graph)



from logging import getLogger

try:
    import cupy as cp
except ImportError:
    logger = getLogger(__name__)
    logger.warning('Cupy is not installed or GPU is not detected.'
                   ' Ignore this message if not using GPU')
    import numpy as cp

import numpy as np
from numba import njit


@njit(nogil=True, fastmath=True)
def logsumexp(a):
    a_max = np.max(a)
    out = np.log(np.sum(np.exp(a - a_max)))
    out += a_max
    return out


@njit(parallel=True, error_model='numpy')
def normalize_to_probability(distribution):
    '''Ensure the distribution integrates to 1 so that it is a probability
    distribution
    '''
    return distribution / np.nansum(distribution)


@njit(nogil=True, error_model='numpy', cache=False)
def _causal_decode(initial_conditions, state_transition, likelihood):
    '''Adaptive filter to iteratively calculate the posterior probability
    of a state variable using past information.

    Parameters
    ----------
    initial_conditions : ndarray, shape (n_bins,)
    state_transition : ndarray, shape (n_bins, n_bins)
    likelihood : ndarray, shape (n_time, n_bins)

    Returns
    -------
    posterior : ndarray, shape (n_time, n_bins)

    '''

    n_time = likelihood.shape[0]
    posterior = np.zeros_like(likelihood)

    posterior[0] = initial_conditions.copy() * likelihood[0]
    norm = np.nansum(posterior[0])
    log_data_likelihood = np.log(norm)
    posterior[0] /= norm

    for k in np.arange(1, n_time):
        posterior[k] = state_transition.T @ posterior[k - 1] * likelihood[k]
        norm = np.nansum(posterior[k])
        log_data_likelihood += np.log(norm)
        posterior[k] /= norm

    return posterior, log_data_likelihood


@njit(nogil=True, error_model='numpy', cache=False)
def _acausal_decode(causal_posterior, state_transition):
    '''Uses past and future information to estimate the state.

    Parameters
    ----------
    causal_posterior : ndarray, shape (n_time, n_bins, 1)
    state_transition : ndarray, shape (n_bins, n_bins)

    Return
    ------
    acausal_posterior : ndarray, shape (n_time, n_bins, 1)

    '''
    acausal_posterior = np.zeros_like(causal_posterior)
    acausal_posterior[-1] = causal_posterior[-1].copy()
    n_time, n_bins = causal_posterior.shape[0], causal_posterior.shape[-2]
    weights = np.zeros((n_bins, 1))

    for time_ind in np.arange(n_time - 2, -1, -1):
        acausal_prior = (
            state_transition.T @ causal_posterior[time_ind])
        log_ratio = (
            np.log(acausal_posterior[time_ind + 1, ..., 0] + np.spacing(1)) -
            np.log(acausal_prior[..., 0] + np.spacing(1)))
        weights[..., 0] = np.exp(log_ratio) @ state_transition

        acausal_posterior[time_ind] = normalize_to_probability(
            weights * causal_posterior[time_ind])

    return acausal_posterior


@njit(nogil=True, error_model='numpy', cache=False)
def _causal_classify(initial_conditions, continuous_state_transition,
                     discrete_state_transition, likelihood):
    '''Adaptive filter to iteratively calculate the posterior probability
    of a state variable using past information.

    Parameters
    ----------
    initial_conditions : ndarray, shape (n_states, n_bins, 1)
    continuous_state_transition : ndarray, shape (n_states, n_states,
                                                  n_bins, n_bins)
    discrete_state_transition : ndarray, shape (n_states, n_states)
    likelihood : ndarray, shape (n_time, n_states, n_bins, 1)

    Returns
    -------
    causal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)

    '''
    n_time, n_states, n_bins, _ = likelihood.shape
    posterior = np.zeros_like(likelihood)

    posterior[0] = initial_conditions.copy() * likelihood[0]
    norm = np.nansum(posterior[0])
    log_data_likelihood = np.log(norm)
    posterior[0] /= norm

    for k in np.arange(1, n_time):
        prior = np.zeros((n_states, n_bins, 1))
        for state_k in np.arange(n_states):
            for state_k_1 in np.arange(n_states):
                prior[state_k, :] += (
                    discrete_state_transition[state_k_1, state_k] *
                    continuous_state_transition[state_k_1, state_k].T @
                    posterior[k - 1, state_k_1])
        posterior[k] = prior * likelihood[k]
        norm = np.nansum(posterior[k])
        log_data_likelihood += np.log(norm)
        posterior[k] /= norm

    return posterior, log_data_likelihood


@njit(nogil=True, error_model='numpy', cache=False)
def _acausal_classify(causal_posterior, continuous_state_transition,
                      discrete_state_transition):
    '''Uses past and future information to estimate the state.

    Parameters
    ----------
    causal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)
    continuous_state_transition : ndarray, shape (n_states, n_states,
                                                  n_bins, n_bins)
    discrete_state_transition : ndarray, shape (n_states, n_states)

    Return
    ------
    acausal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)

    '''
    acausal_posterior = np.zeros_like(causal_posterior)
    acausal_posterior[-1] = causal_posterior[-1].copy()
    n_time, n_states, n_bins, _ = causal_posterior.shape

    for k in np.arange(n_time - 2, -1, -1):
        # Prediction Step -- p(x_{k+1}, I_{k+1} | y_{1:k})
        prior = np.zeros((n_states, n_bins, 1))
        for state_k_1 in np.arange(n_states):
            for state_k in np.arange(n_states):
                prior[state_k_1, :] += (
                    discrete_state_transition[state_k, state_k_1] *
                    continuous_state_transition[state_k, state_k_1].T @
                    causal_posterior[k, state_k])

        # Backwards Update
        weights = np.zeros((n_states, n_bins, 1))
        ratio = np.exp(
            np.log(acausal_posterior[k + 1] + np.spacing(1)) -
            np.log(prior + np.spacing(1)))
        for state_k in np.arange(n_states):
            for state_k_1 in np.arange(n_states):
                weights[state_k] += (
                    discrete_state_transition[state_k, state_k_1] *
                    continuous_state_transition[state_k, state_k_1] @
                    ratio[state_k_1])

        acausal_posterior[k] = normalize_to_probability(
            weights * causal_posterior[k])

    return acausal_posterior


def scaled_likelihood(log_likelihood, axis=1):
    '''
    Parameters
    ----------
    log_likelihood : ndarray, shape (n_time, n_bins)

    Returns
    -------
    scaled_log_likelihood : ndarray, shape (n_time, n_bins)

    '''
    max_log_likelihood = np.nanmax(log_likelihood, axis=axis, keepdims=True)
    # If maximum is infinity, set to zero
    if max_log_likelihood.ndim > 0:
        max_log_likelihood[~np.isfinite(max_log_likelihood)] = 0.0
    elif not np.isfinite(max_log_likelihood):
        max_log_likelihood = 0.0

    # Maximum likelihood is always 1
    likelihood = np.exp(log_likelihood - max_log_likelihood)
    likelihood += np.spacing(1)  # avoid zero likelihood
    return likelihood


def mask(value, is_track_interior):
    try:
        value[..., ~is_track_interior] = np.nan
    except IndexError:
        value[..., ~is_track_interior, :] = np.nan
    return value


def normalize_to_probability_gpu(distribution):
    '''Ensure the distribution integrates to 1 so that it is a probability
    distribution
    '''
    return distribution / cp.nansum(distribution)


def _causal_decode_gpu(initial_conditions, state_transition, likelihood):
    '''Adaptive filter to iteratively calculate the posterior probability
    of a state variable using past information.

    Parameters
    ----------
    initial_conditions : ndarray, shape (n_bins,)
    state_transition : ndarray, shape (n_bins, n_bins)
    likelihood : ndarray, shape (n_time, n_bins)

    Returns
    -------
    posterior : ndarray, shape (n_time, n_bins)

    '''

    initial_conditions = cp.asarray(initial_conditions, dtype=cp.float32)
    state_transition = cp.asarray(state_transition, dtype=cp.float32)
    likelihood = cp.asarray(likelihood, dtype=cp.float32)

    n_time = likelihood.shape[0]
    posterior = cp.zeros_like(likelihood)

    posterior[0] = initial_conditions * likelihood[0]
    norm = cp.nansum(posterior[0])
    log_data_likelihood = cp.log(norm)
    posterior[0] /= norm

    for k in np.arange(1, n_time):
        posterior[k] = state_transition.T @ posterior[k - 1] * likelihood[k]
        norm = np.nansum(posterior[k])
        log_data_likelihood += cp.log(norm)
        posterior[k] /= norm

    return cp.asnumpy(posterior), cp.asnumpy(log_data_likelihood)


def _acausal_decode_gpu(causal_posterior, state_transition):
    '''Uses past and future information to estimate the state.

    Parameters
    ----------
    causal_posterior : ndarray, shape (n_time, n_bins, 1)
    state_transition : ndarray, shape (n_bins, n_bins)

    Return
    ------
    acausal_posterior : ndarray, shape (n_time, n_bins, 1)

    '''
    causal_posterior = cp.asarray(causal_posterior, dtype=cp.float32)
    state_transition = cp.asarray(state_transition, dtype=cp.float32)

    acausal_posterior = cp.zeros_like(causal_posterior)
    acausal_posterior[-1] = causal_posterior[-1]
    n_time, n_bins = causal_posterior.shape[0], causal_posterior.shape[-2]
    weights = cp.zeros((n_bins, 1))

    for time_ind in np.arange(n_time - 2, -1, -1):
        acausal_prior = (
            state_transition.T @ causal_posterior[time_ind])
        log_ratio = (
            cp.log(acausal_posterior[time_ind + 1, ..., 0] + np.spacing(1)) -
            cp.log(acausal_prior[..., 0] + np.spacing(1)))
        weights[..., 0] = cp.exp(log_ratio) @ state_transition

        acausal_posterior[time_ind] = normalize_to_probability_gpu(
            weights * causal_posterior[time_ind])

    return cp.asnumpy(acausal_posterior)


def _causal_classify_gpu(initial_conditions, continuous_state_transition,
                         discrete_state_transition, likelihood):
    '''Adaptive filter to iteratively calculate the posterior probability
    of a state variable using past information.

    Parameters
    ----------
    initial_conditions : ndarray, shape (n_states, n_bins, 1)
    continuous_state_transition : ndarray, shape (n_states, n_states,
                                                  n_bins, n_bins)
    discrete_state_transition : ndarray, shape (n_states, n_states)
    likelihood : ndarray, shape (n_time, n_states, n_bins, 1)

    Returns
    -------
    causal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)

    '''
    initial_conditions = cp.asarray(initial_conditions, dtype=cp.float32)
    continuous_state_transition = cp.asarray(
        continuous_state_transition, dtype=cp.float32)
    discrete_state_transition = cp.asarray(
        discrete_state_transition, dtype=cp.float32)
    likelihood = cp.asarray(likelihood, dtype=cp.float32)

    n_time, n_states, n_bins, _ = likelihood.shape
    posterior = cp.zeros_like(likelihood)

    posterior[0] = initial_conditions * likelihood[0]
    norm = cp.nansum(posterior[0])
    log_data_likelihood = cp.log(norm)
    posterior[0] /= norm

    for k in np.arange(1, n_time):
        prior = cp.zeros((n_states, n_bins, 1))
        for state_k in np.arange(n_states):
            for state_k_1 in np.arange(n_states):
                prior[state_k, :] += (
                    discrete_state_transition[state_k_1, state_k] *
                    continuous_state_transition[state_k_1, state_k].T @
                    posterior[k - 1, state_k_1])
        posterior[k] = prior * likelihood[k]
        norm = cp.nansum(posterior[k])
        log_data_likelihood += cp.log(norm)
        posterior[k] /= norm

    return cp.asnumpy(posterior), cp.asnumpy(log_data_likelihood)


def _acausal_classify_gpu(causal_posterior, continuous_state_transition,
                          discrete_state_transition):
    '''Uses past and future information to estimate the state.

    Parameters
    ----------
    causal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)
    continuous_state_transition : ndarray, shape (n_states, n_states,
                                                  n_bins, n_bins)
    discrete_state_transition : ndarray, shape (n_states, n_states)

    Return
    ------
    acausal_posterior : ndarray, shape (n_time, n_states, n_bins, 1)

    '''
    causal_posterior = cp.asarray(causal_posterior, dtype=cp.float32)
    continuous_state_transition = cp.asarray(
        continuous_state_transition, dtype=cp.float32)
    discrete_state_transition = cp.asarray(
        discrete_state_transition, dtype=cp.float32)

    acausal_posterior = cp.zeros_like(causal_posterior)
    acausal_posterior[-1] = causal_posterior[-1]
    n_time, n_states, n_bins, _ = causal_posterior.shape

    for k in np.arange(n_time - 2, -1, -1):
        # Prediction Step -- p(x_{k+1}, I_{k+1} | y_{1:k})
        prior = cp.zeros((n_states, n_bins, 1))
        for state_k_1 in np.arange(n_states):
            for state_k in np.arange(n_states):
                prior[state_k_1, :] += (
                    discrete_state_transition[state_k, state_k_1] *
                    continuous_state_transition[state_k, state_k_1].T @
                    causal_posterior[k, state_k])

        # Backwards Update
        weights = cp.zeros((n_states, n_bins, 1))
        ratio = cp.exp(
            cp.log(acausal_posterior[k + 1] + np.spacing(1)) -
            cp.log(prior + np.spacing(1)))
        for state_k in np.arange(n_states):
            for state_k_1 in np.arange(n_states):
                weights[state_k] += (
                    discrete_state_transition[state_k, state_k_1] *
                    continuous_state_transition[state_k, state_k_1] @
                    ratio[state_k_1])

        acausal_posterior[k] = normalize_to_probability_gpu(
            weights * causal_posterior[k])

    return cp.asnumpy(acausal_posterior)


def check_converged(loglik, previous_loglik, tolerance=1e-4):
    delta_loglik = abs(loglik - previous_loglik)
    avg_loglik = (abs(loglik) + abs(previous_loglik) + np.spacing(1)) / 2

    is_increasing = loglik - previous_loglik >= -1e-3
    is_converged = (delta_loglik / avg_loglik) < tolerance

    return is_converged, is_increasing



import copy
import math

import dask.bag as db
import networkx as nx
import numba
import numpy as np

SQRT_2PI = np.sqrt(2.0 * np.pi)


@numba.vectorize(['float64(float64, float64)'], nopython=True,
                 cache=True)
def gaussian_kernel(distance, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * (distance / sigma)**2) / (sigma * SQRT_2PI)


def _distance_to_bin_centers(left_node, right_node, distance_left_node,
                             distance_right_node, time_ind, copy_graph,
                             place_bin_center_node_ids):
    node_name = f"actual_position_{time_ind}"
    copy_graph.add_node(node_name)
    copy_graph.add_edge(
        left_node, node_name, distance=distance_left_node)
    copy_graph.add_edge(
        node_name, right_node, distance=distance_right_node)
    distance_to_bin_centers = [
        nx.shortest_path_length(
            copy_graph, source=bin_center, target=node_name,
            weight="distance")
        for bin_center in place_bin_center_node_ids]
    copy_graph.remove_node(node_name)

    return distance_to_bin_centers


def _find_adjacent_nodes(nodes_df, linear_position):
    # Find the index of the nodes to insert between
    right_bin_ind = np.searchsorted(
        nodes_df.linear_position.values, linear_position, side="right")
    left_bin_ind = right_bin_ind - 1

    # Fix indices that are exactly the maximum bin edge
    n_bins = nodes_df.linear_position.shape[0]
    right_bin_ind[right_bin_ind >= n_bins] -= 1

    # Fix for ones that fall into invalid track positions
    # Note: Need to test on different binnings
    # Another solution is to find the edge this falls on directly.
    not_same_edge = (nodes_df.edge_id.values[left_bin_ind] !=
                     nodes_df.edge_id.values[right_bin_ind])
    right_bin_ind[not_same_edge] = right_bin_ind[not_same_edge] - 1
    left_bin_ind[not_same_edge] = left_bin_ind[not_same_edge] - 1

    # Get adjacent node names and distance
    left_node = nodes_df.reset_index().node_id.values[left_bin_ind]
    right_node = nodes_df.reset_index().node_id.values[right_bin_ind]

    distance_left_node = np.abs(
        nodes_df.loc[left_node].linear_position.values - linear_position)
    distance_right_node = np.abs(
        nodes_df.loc[right_node].linear_position.values - linear_position)

    return left_node, right_node, distance_left_node, distance_right_node


def get_distance_to_bin_centers(linear_position, decoder, npartitions=100):
    copy_graph = copy.deepcopy(decoder.track_graph_with_bin_centers_edges_)
    linear_position = linear_position.squeeze()
    nodes_df = decoder.nodes_df_.set_index("node_id")
    place_bin_center_node_ids = (
        nodes_df
        .loc[~nodes_df.is_bin_edge]
        .reset_index()
        .node_id
        .values)
    (left_node, right_node, distance_left_node,
     distance_right_node) = _find_adjacent_nodes(nodes_df, linear_position)

    left_node = db.from_sequence(
        left_node, npartitions=npartitions)
    right_node = db.from_sequence(
        right_node, npartitions=npartitions)
    distance_left_node = db.from_sequence(
        distance_left_node, npartitions=npartitions)
    distance_right_node = db.from_sequence(
        distance_right_node, npartitions=npartitions)
    time_ind = db.from_sequence(
        range(linear_position.shape[0]), npartitions=npartitions)

    return np.asarray(left_node.map(
        _distance_to_bin_centers, right_node, distance_left_node,
        distance_right_node, time_ind, copy_graph,
        place_bin_center_node_ids).compute())



#!/usr/bin/env python3

from setuptools import find_packages, setup

INSTALL_REQUIRES = ['numpy', 'statsmodels', 'numba', 'matplotlib',
                    'pandas', 'xarray', 'scipy', 'scikit-learn', 'scikit-image',
                    'regularized_glm', 'dask', 'patsy', 'networkx', 'joblib',
                    'track_linearization', 'tqdm']
TESTS_REQUIRE = ['pytest >= 2.7.1']

setup(
    name='replay_trajectory_classification',
    version='1.0.16',
    license='MIT',
    description=('Classify replay trajectories.'),
    author='Eric Denovellis',
    author_email='eric.denovellis@ucsf.edu',
    url='https://github.com/Eden-Kramer-Lab/replay_trajectory_classification',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
)



from dataclasses import dataclass

import numpy as np


@dataclass
class UniformInitialConditions:
    def make_initial_conditions(self,
                                environments: tuple,
                                environment_names_to_state: tuple):
        n_total_place_bins = 0
        initial_conditions = []

        for environment_name in environment_names_to_state:
            is_track_interior = (
                environments[environments.index(environment_name)]
                .is_track_interior_
                .ravel(order='F'))
            n_total_place_bins += is_track_interior.sum()
            initial_conditions.append(is_track_interior)

        return [ic / n_total_place_bins for ic in initial_conditions]


@dataclass
class UniformOneEnvironmentInitialConditions:
    environment_name: str = ''

    def make_initial_conditions(self,
                                environments: tuple,
                                environment_names_to_state: tuple):
        n_total_place_bins = 0
        initial_conditions = []

        for environment_name in environment_names_to_state:
            is_track_interior = (
                environments[environments.index(environment_name)]
                .is_track_interior_
                .ravel(order='F'))

            if self.environment_name == environment_name:
                initial_conditions.append(is_track_interior)
                n_total_place_bins += is_track_interior.sum()
            else:
                initial_conditions.append(np.zeros_like(is_track_interior))

        return [ic / n_total_place_bins for ic in initial_conditions]



import numpy as np
from replay_trajectory_classification.bins import atleast_2d
from scipy.stats import multivariate_normal


def _normalize_row_probability(x):
    '''Ensure the state transition matrix rows sum to 1
    '''
    x /= x.sum(axis=1, keepdims=True)
    x[np.isnan(x)] = 0
    return x


def empirical_movement(
    place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes):
    '''Estimate the probablity of the next position based on the movement
     data, given the movment is sped up by the
     `replay_speed`

    Place cell firing during a hippocampal replay event is a "sped-up"
    version of place cell firing when the animal is actually moving.
    Here we use the animal's actual movements to constrain which place
    cell is likely to fire next.

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    edges : sequence
        A sequence of arrays describing the bin edges along each dimension.
    is_training : None or bool ndarray, shape (n_time,), optional
    replay_speed : int, optional
        How much the movement is sped-up during a replay event
    position_extent : sequence, optional
        A sequence of `n_position_dims`, each an optional (lower, upper)
        tuple giving the outer bin edges for position.
        An entry of None in the sequence results in the minimum and maximum
        values being used for the corresponding dimension.
        The default, None, is equivalent to passing a tuple of
        `n_position_dims` None values.

    Returns
    -------
    transition_matrix : ndarray, shape (n_position_bins, n_position_bins)

    '''
    if is_training is None:
        is_training = np.ones((position.shape[0]), dtype=np.bool)
    position = atleast_2d(position)[is_training]
    movement_bins, _ = np.histogramdd(
        np.concatenate((position[1:], position[:-1]), axis=1),
        bins=edges * 2, range=position_extent)
    original_shape = movement_bins.shape
    n_position_dims = position.shape[1]
    shape_2d = np.product(original_shape[:n_position_dims])
    movement_bins = _normalize_row_probability(
        movement_bins.reshape((shape_2d, shape_2d), order='F'))
    movement_bins = np.linalg.matrix_power(movement_bins, replay_speed)

    return movement_bins


def random_walk(place_bin_centers, is_track_interior, position, edges,
                is_training, replay_speed, position_extent, movement_var,
                place_bin_center_ind_to_node,
                distance_between_nodes):
    '''Zero mean random walk.

    This version makes the sped up gaussian without constraints and then
    imposes the constraints.

    Parameters
    ----------
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    movement_var : float,
    is_track_interior : bool ndarray, shape (n_x_bins, n_y_bins)
    replay_speed : int

    Returns
    -------
    transition_matrix : ndarray, shape (n_bins, n_bins)

    '''
    if place_bin_center_ind_to_node is None:
        transition_matrix = np.stack(
            [multivariate_normal(
                mean=bin, cov=movement_var * replay_speed).pdf(
                place_bin_centers)
             for bin in place_bin_centers], axis=1)
    else:
        transition_matrix = _random_walk_on_track_graph(
            place_bin_centers, is_track_interior, position, edges,
            is_training, replay_speed, position_extent, movement_var,
            place_bin_center_ind_to_node,
            distance_between_nodes,
        )
    is_track_interior = is_track_interior.ravel(order='F')
    transition_matrix[~is_track_interior] = 0.0
    transition_matrix[:, ~is_track_interior] = 0.0

    return _normalize_row_probability(transition_matrix)


def _random_walk_on_track_graph(
    place_bin_centers, is_track_interior, position, edges,
    is_training, replay_speed, position_extent, movement_var,
    place_bin_center_ind_to_node,
    distance_between_nodes
):
    """

    Parameters
    ----------
    node_linear_position : np.ndarray, shape (n_nodes,)
    place_bin_centers : np.ndarray, shape (n_bins)
    is_track_interior : np.ndarray, shape (n_bins)
    distance_between_nodes : dict of dicts, shape (n_nodes, n_nodes)
    bin_to_node : tuple, shape (n_nodes,)
    movement_var : float, optional
    replay_speed : float, optional

    Returns
    -------
    state_transition : np.ndarray, shape (n_bins, n_bins)
    bin_to_node : np.array, shape (n_bins)

    """
    state_transition = np.zeros(
        (place_bin_centers.size, place_bin_centers.size))
    gaussian = multivariate_normal(mean=0, cov=movement_var * replay_speed)

    for bin_ind1, node1 in enumerate(place_bin_center_ind_to_node):
        for bin_ind2, node2 in enumerate(place_bin_center_ind_to_node):
            try:
                state_transition[bin_ind1, bin_ind2] = gaussian.pdf(
                    distance_between_nodes[node1][node2]
                )
            except KeyError:
                # bins not on track interior will be -1 and not in distance
                # between nodes
                continue

    return state_transition


def random_walk_direction1(*args):
    '''Zero mean random walk.

    This version makes the sped up gaussian without constraints and then
    imposes the constraints.

    Parameters
    ----------
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    movement_var : float,
    is_track_interior : bool ndarray, shape (n_x_bins, n_y_bins)
    replay_speed : int

    Returns
    -------
    transition_matrix : ndarray, shape (n_bins, n_bins)

    '''
    random = random_walk(*args)
    return _normalize_row_probability(np.triu(random))


def random_walk_direction2(*args):
    '''Zero mean random walk.

    This version makes the sped up gaussian without constraints and then
    imposes the constraints.

    Parameters
    ----------
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    movement_var : float,
    is_track_interior : bool ndarray, shape (n_x_bins, n_y_bins)
    replay_speed : int

    Returns
    -------
    transition_matrix : ndarray, shape (n_bins, n_bins)

    '''
    random = random_walk(*args)
    return _normalize_row_probability(np.tril(random))


def uniform_state_transition(
    place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes):
    '''Equally likely to go somewhere on the track.

    Parameters
    ----------
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    is_track_interior : bool ndarray, shape (n_x_bins, n_y_bins)

    Returns
    -------
    transition_matrix : ndarray, shape (n_bins, n_bins)

    '''
    n_bins = place_bin_centers.shape[0]
    transition_matrix = np.ones((n_bins, n_bins))

    is_track_interior = is_track_interior.ravel(order='F')
    transition_matrix[~is_track_interior] = 0.0
    transition_matrix[:, ~is_track_interior] = 0.0

    return _normalize_row_probability(transition_matrix)


def identity(place_bin_centers, is_track_interior, position, edges,
             is_training, replay_speed, position_extent, movement_var,
             place_bin_center_ind_to_node,
             distance_between_nodes):
    '''Stay in one place on the track.

    Parameters
    ----------
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    is_track_interior : bool ndarray, shape (n_x_bins, n_y_bins)

    Returns
    -------
    transition_matrix : ndarray, shape (n_bins, n_bins)

    '''
    n_bins = place_bin_centers.shape[0]
    transition_matrix = np.identity(n_bins)

    is_track_interior = is_track_interior.ravel(order='F')
    transition_matrix[~is_track_interior] = 0.0
    transition_matrix[:, ~is_track_interior] = 0.0

    return _normalize_row_probability(transition_matrix)


def uniform_minus_empirical(*args):
    uniform = uniform_state_transition(*args)
    empirical = empirical_movement(*args)
    difference = uniform - empirical
    difference[difference < 0] = 0.0
    return _normalize_row_probability(difference)


def uniform_minus_random_walk(*args):
    uniform = uniform_state_transition(*args)
    random = random_walk(*args)
    difference = uniform - random
    difference[difference < 0] = 0.0
    return _normalize_row_probability(difference)


def empirical_minus_identity(*args):
    empirical = empirical_movement(*args)
    ident = identity(*args)
    difference = empirical - ident
    difference[difference < 0] = 0.0
    return _normalize_row_probability(difference)


def random_walk_minus_identity(*args):
    random = random_walk(*args)
    ident = identity(*args)
    difference = random - ident
    difference[difference < 0] = 0.0
    return _normalize_row_probability(difference)


def inverse_random_walk(
    place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes):
    rw = random_walk(
        place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes)
    transition_matrix = rw.max(axis=1, keepdims=True) - rw

    is_track_interior = is_track_interior.ravel(order='F')
    transition_matrix[~is_track_interior] = 0.0
    transition_matrix[:, ~is_track_interior] = 0.0
    return _normalize_row_probability(transition_matrix)


def random_walk_plus_uniform(
    place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes):
    random = random_walk(
        place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes)
    uniform = uniform_state_transition(
        place_bin_centers, is_track_interior, position, edges,
        is_training, replay_speed, position_extent, movement_var,
        place_bin_center_ind_to_node,
        distance_between_nodes)
    transition_matrix = uniform + random
    is_track_interior = is_track_interior.ravel(order='F')
    transition_matrix[~is_track_interior] = 0.0
    transition_matrix[:, ~is_track_interior] = 0.0

    return _normalize_row_probability(transition_matrix)


def identity_discrete(n_states, diag):
    '''

    Parameters
    ----------
    n_states : int

    Returns
    -------
    transition_matrix : ndarray, shape (n_states, n_states)

    '''
    return np.identity(n_states)


def strong_diagonal_discrete(n_states, diag):
    '''

    Parameters
    ----------
    n_states : int
    diag : float

    Returns
    -------
    transition_matrix : ndarray, shape (n_states, n_states)

    '''
    strong_diagonal = np.identity(n_states) * diag
    is_off_diag = ~np.identity(n_states, dtype=bool)
    strong_diagonal[is_off_diag] = (
        (1 - diag) / (n_states - 1))
    return strong_diagonal


def uniform_discrete(n_states, diag):
    '''

    Parameters
    ----------
    n_states : int

    Returns
    -------
    transition_matrix : ndarray, shape (n_states, n_states)

    '''
    return np.ones((n_states, n_states)) / n_states


def estimate_movement_var(position, sampling_frequency):
    '''Estimates the movement variance based on position.

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dim)

    Returns
    -------
    movement_std : ndarray, shape (n_position_dim,)

    '''
    position = atleast_2d(position)
    is_nan = np.any(np.isnan(position), axis=1)
    position = position[~is_nan]
    movement_var = np.cov(np.diff(position, axis=0), rowvar=False)
    return movement_var * sampling_frequency


CONTINUOUS_TRANSITIONS = {
    'empirical_movement': empirical_movement,
    'random_walk': random_walk,
    'uniform': uniform_state_transition,
    'identity': identity,
    'uniform_minus_empirical': uniform_minus_empirical,
    'uniform_minus_random_walk': uniform_minus_random_walk,
    'empirical_minus_identity': empirical_minus_identity,
    'random_walk_minus_identity': random_walk_minus_identity,
    'random_walk_plus_uniform': random_walk_plus_uniform,
    'inverse_random_walk': inverse_random_walk,
    'random_walk_direction1': random_walk_direction1,
    'random_walk_direction2': random_walk_direction2,
}

DISCRETE_TRANSITIONS = {
    'strong_diagonal': strong_diagonal_discrete,
    'identity': identity_discrete,
    'uniform': uniform_discrete,
}
